///////////////////////////////////////////////////////////////////
// FILE:      main.cpp
// CONTAINS:  template code for CPSC 426, Feb 2014, Assignment 2
///////////////////////////////////////////////////////////////////

#include "defs.h"
#include "TriangleMesh.h"
#include "FreeFormDeformation.h"

#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>

// Camera parameters
int windowWidth = 1024;         // window width (pixels)
int windowHeight = 768;         // window height (pixels)
float cameraFovy = 35.0;        // field of view  in the y-direction (degrees)
float cameraZnear = 1.0;        // near clipping plane (distance from camera)
float cameraZfar = 100.0;       // far clipping plane (distance from camera)
float cameraR = 30;             // camera position: distance to center (OpenGL units)
float cameraTheta = 0.25*3.14;  // camera position: angle theta (radians)
float cameraPhi = 0.25*3.14;    // camera position: angle phi(radians)
float cameraCenter[] = {0,0,0}; // where the camera is looking at (OpenGL units)
float cameraUp[] = {0,1,0};     // up vector of the camera (OpenGL units)

// Rendering parameters
FFDControlPoints::DrawStyle latticeDrawStyle = FFDControlPoints::Lattice;
TriangleMesh::MeshDrawStyle meshDrawStyle = TriangleMesh::SOLID;

// Animation parameters
bool paused = false;   // is the animation paused?
float currentTime = 0; // current time of the animation (seconds)
float maxTime = 3.0;   // maximum time (seconds). 'time' loops from 0 to maxTime
float deltaTime = 0.0; // elapsed time since last update (seconds)
int elapsedTime = 0;   // elapsed time since program started (milliseconds)
float dtScale = 1.0;   // scaling factor for timesteps in physical simulation

// Default scene file  and   scenario
string sceneFile("scene.txt");
int mode = 0;

// 
TriangleMesh mesh;
Affine3 bindFrame;
FFDControlPoints bindControlPoints;
FFDControlPoints tempControlPoints;
FFDControlPoints offsetControlPoints;
FFDControlPoints deformedControlPoints;
vector<FFDControlPoints> keyframeControlPoints;

extern void initScene();
extern void updateScene();
extern void loadScene();
extern void computeFFDCoords(TriangleMesh &mesh, Affine3 &frame);
extern void physicsInit();
extern void physicsUpdate();
extern void computeDeformedMesh();

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
//@               TO IMPLEMENT               @//
//        Add your global variables here

//@                                          @//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//

///////////////////////////////////////////////////////////////////
// FUNC:  init()
// DOES:  initializes default light source and surface material
///////////////////////////////////////////////////////////////////

void init(void) 
{
   glClearColor (0.8f, 0.8f, 0.8f, 1.0);
   glShadeModel (GL_SMOOTH);
   glEnable(GL_LIGHTING);
   glEnable(GL_LIGHT0);
   glEnable(GL_DEPTH_TEST);
   glDepthFunc(GL_LEQUAL);
   glEnable(GL_COLOR_MATERIAL);
   glEnable(GL_NORMALIZE);

   initScene();
   updateScene();
}

///////////////////////////////////////////////////////////////////
// FUNC: initScene()
// DOES: initialize the scene
///////////////////////////////////////////////////////////////////

void initScene()
{
  currentTime = 0;     // reset time
  maxTime = 3.0;       // default animation time

  loadScene();         // read scene description:  mesh, binding frame, target keyframe
  computeFFDCoords(mesh, bindFrame);  // compute s,t,u coords for each vertex

  switch(mode) {   // mode-specific initialization
  case 0:
    bindControlPoints = FFDControlPoints(bindFrame);  // initial control point locations
    break;
  case 1:
    break;
  case 2:
    break;
  case 3:
  case 4:
    physicsInit();
    break;
  }
}

///////////////////////////////////////////////////////////////////
// FUNC: computeFFDCoords()
// DOES: update the location of all objects/vertices in the scene, as a function of Time
///////////////////////////////////////////////////////////////////

void computeFFDCoords(TriangleMesh &mesh, Affine3 &bindFrame)
{
  //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
  //@               TO IMPLEMENT               @//
  //     Compute the FFD vertex coordinates

  //@                                          @//
  //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
}

///////////////////////////////////////////////////////////////////
// FUNC: computeDeformedMesh()
// DOES: compute new location of all vertices based on the current deformed lattice
///////////////////////////////////////////////////////////////////

void computeDeformedMesh()
{
  //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
  //@               TO IMPLEMENT               @//
  //    Deform the mesh according using the
  //    pre-computed FFD coordinates and the
  //    control points 'deformedControlPoints'

}

///////////////////////////////////////////////////////////////////
// FUNC: updateScene()
// DOES: update the location of all objects/vertices in the scene, as a function of Time
///////////////////////////////////////////////////////////////////

void updateScene()
{
  double u;
  Affine3 currentBindFrame;
  Vector3 orig;

  switch(mode) {
  case 0:
    deformedControlPoints = bindControlPoints;    // no motion
    break;
  case 1:         // interpolate control points between the first two keyframes
    u = currentTime < 0.5*maxTime ? 2*currentTime/maxTime : 2*(maxTime-currentTime)/maxTime;
    deformedControlPoints = linearInterpolation(u, keyframeControlPoints[0], keyframeControlPoints[1]);
    break;
  case 2:         // interpolate control points between the first two keyframes, with rebinding of lattice
    u = currentTime < 0.5*maxTime ? 2*currentTime/maxTime : 2*(maxTime-currentTime)/maxTime;
    deformedControlPoints = linearInterpolation(u, keyframeControlPoints[0], keyframeControlPoints[1]);
    currentBindFrame = Affine3(bindFrame.getcol(0), bindFrame.getcol(1), bindFrame.getcol(2), deformedControlPoints(0,0,0));
    break;
  case 3:
  case 4:
    physicsUpdate();
    break;
  }
  computeDeformedMesh();     // now compute the deformed mesh
}

///////////////////////////////////////////////////////////////////
// FUNC: drawScene()
// DOES: draw the scene as updated by the function updateScene()
///////////////////////////////////////////////////////////////////

void drawScene()
{
  // Draw ground plane
  glColor3f(0.5,0.5,0.9);
  glNormal3f(0,1,0);
  glBegin(GL_POLYGON);
  {
    glVertex3f(-10,-0.101,-10);
    glVertex3f(-10,-0.101,10);
    glVertex3f(10,-0.101,10);
    glVertex3f(10,-0.101,-10);
  }
  glEnd();
		
  // Draw control points
  deformedControlPoints.draw(latticeDrawStyle);

  // Draw mesh
  glColor3f(0.7,0.5,0.1);   // brownish color
  mesh.draw(meshDrawStyle);
}

///////////////////////////////////////////////////////////////////
// FUNC:  displayCallback()
// DOES:  clears screen, draws the world, swaps buffers
///////////////////////////////////////////////////////////////////

void displayCallback(void)
{
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   // clear draw buffer
  glMatrixMode(GL_MODELVIEW);  // setup ModelView matrix for the desired camera position
  glLoadIdentity();
  float eye [] = 	{ 
		cameraCenter[0] + cameraR * std::sin(cameraTheta) * std::sin(cameraPhi),
		cameraCenter[1] + cameraR * std::cos(cameraTheta),
		cameraCenter[2] + cameraR * std::sin(cameraTheta) * std::cos(cameraPhi)  	};
  gluLookAt(eye[0],eye[1],eye[2],cameraCenter[0],cameraCenter[1],cameraCenter[2],cameraUp[0],cameraUp[1], cameraUp[2]);
  drawScene();        // draw scene
  glutSwapBuffers();  // swap buffers
}

///////////////////////////////////////////////////////////////////
// FUNC:  keyCallback()
// DOES:  handles glut keyboard events
///////////////////////////////////////////////////////////////////

void keyCallback(unsigned char key, int x, int y)
{
  unsigned int i,n;

  if( (key >= '0') && (key <= '9') ) {
    mode = (int) key - (int) '0';
    initScene();
    updateScene();
  }
  switch(key) {
  case 'q': 
    exit(0);
    break;
  case 'r':
    currentTime = 0; // set global time to zero
    paused = false;  // run animation
    break;
  case ' ':
    paused = !paused;
    break;
  case 'l':
    ++latticeDrawStyle; // cycle through draw styles
    break;
  case 'm':
    ++meshDrawStyle;   // cycle through mesh draw styles
    break;
  case 'p':
    n = keyframeControlPoints.size();
    for (i=0; i<n; i++) {
      keyframeControlPoints[i].print("keyframe ");
    }
    mesh.print();
    break;
  default:
    break;
  }
  glutPostRedisplay();
}

///////////////////////////////////////////////////////////////////
// FUNC:  specialKeyCallback()
// DOES:  called for non-ASCII key strokes, e.g., arrow keys
///////////////////////////////////////////////////////////////////

void specialKeyCallback(int key, int x, int y)
{
  switch(key) {
  case GLUT_KEY_RIGHT: 
    cameraPhi += 3.14 / 16;
    break;
  case GLUT_KEY_LEFT: 
    cameraPhi -= 3.14 / 16;
    break;
  case GLUT_KEY_UP: 
    cameraTheta -= 3.14 / 16;
    if(cameraTheta < 0.01)
      cameraTheta = 0.01;
    break;
  case GLUT_KEY_DOWN: 
    cameraTheta += 3.14 / 16;
    if(cameraTheta > 3.14)
      cameraTheta = 3.14;
    break;
  case GLUT_KEY_PAGE_UP: 
    cameraR /= 1.2;
    break;
  case GLUT_KEY_PAGE_DOWN: 
    cameraR *= 1.2;
    break;
  default:
    break;
  }
  glutPostRedisplay();
}

///////////////////////////////////////////////////////////////////
// FUNC:  myReshape()
// DOES:  this is called if the on-screen window is resized
///////////////////////////////////////////////////////////////////

void myReshape(int w, int h)
{
   windowWidth = w;
   windowHeight = h;
   glViewport(0, 0, w, h);
   glMatrixMode(GL_PROJECTION); 
   glLoadIdentity();
   gluPerspective(cameraFovy, (GLfloat) w/h, 1.0, 10000.0);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
}

///////////////////////////////////////////////////////////////////
// FUNC:  idleCallback()
// DOES:  glut calls this when idle;  used for animation
///////////////////////////////////////////////////////////////////

void idleCallback()
{
  if(paused)    {
    elapsedTime = glutGet(GLUT_ELAPSED_TIME);        // Update elapsed time
    return;
  }
  int newElapsedTime = glutGet(GLUT_ELAPSED_TIME);  // Compute new time and update elapsed time
  deltaTime = 0.001*dtScale*(newElapsedTime - elapsedTime);
  elapsedTime = newElapsedTime;
  currentTime += deltaTime;
  if (currentTime>=maxTime)   currentTime = 0.0;
  updateScene();         // update scene
  glutPostRedisplay();   // draw scene
}

///////////////////////////////////////////////////////////////////
// FUNC:  main()
// DOES:  intializes glut window, calls init(), then hands over control to glut
///////////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{
  if (argc==2)
    sceneFile = argv[1];
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
   glutInitWindowSize(windowWidth, windowHeight);
   glutCreateWindow("CPSC 426 Jan 2014:  Assmt 2");
   glutReshapeFunc(myReshape);
   glutDisplayFunc(displayCallback);
   glutKeyboardFunc(keyCallback);
   glutSpecialFunc(specialKeyCallback);
   glutIdleFunc(idleCallback);
   init();
   glutMainLoop();
   return 0;
}

///////////////////////////////////////////////////////////////////
// FUNC: getParsedLine()
// DOES: reads in a line and parses it into words;  also discards empty lines and comment lines
///////////////////////////////////////////////////////////////////

int getParsedLine(ifstream &file, vector<string>  &words)
{
  using namespace std;
  string line;

  words.clear();
  while(getline(file,line)) {               // for each line
    // Split the line into words
    string buf;
    stringstream ss(line);
    while (ss >> buf)
      words.push_back(buf);
    if(words.empty())
      continue;
    if (words[0][0] == '#') {        // comment lines:  discard
      words.clear();
      continue; 
    }
    break;
  }
  return words.size();
  //  cout << words.size() << " words read." << endl;
}

bool absFrame = false;
bool keyframeStart = false;
extern void kfEnd();

///////////////////////////////////////////////////////////////////
// FUNC: kfStart(words)
// DOES: start of a keyframe
///////////////////////////////////////////////////////////////////

void kfStart(vector<string> &words) {
  Affine3 tempFrame;

  if (keyframeStart)     // end the previous keyframe specification if needed
    kfEnd();
  absFrame = false;
  keyframeStart = true;
  offsetControlPoints.zero();
  if (words.size()==1) {           // no affine frame? Assume that absolute x,y,z control points will follow
    absFrame = true;
    cout << "abs frame" << endl;
    return;
  }
  double val[12];                  //  read in frame info
  for (int n=0; n<12; n++) 
    val[n] = atof(words[n+1].c_str());   // read in 12 numbers
  tempFrame = Affine3(val);              // remember the new frame
  tempControlPoints = FFDControlPoints(tempFrame);
}

///////////////////////////////////////////////////////////////////
// FUNC: kfControlPoint()
// DOES: keyframe control point
///////////////////////////////////////////////////////////////////

void kfControlPoint(vector<string> &words) {
  int i,j,k;
  sscanf(words[0].c_str(),"(%d,%d,%d):",&i,&j,&k);
  offsetControlPoints(i,j,k)[0] = atof(words[1].c_str());
  offsetControlPoints(i,j,k)[1] = atof(words[2].c_str());
  offsetControlPoints(i,j,k)[2] = atof(words[3].c_str());
}

///////////////////////////////////////////////////////////////////
// FUNC: kfEnd()
// DOES: end of a keyframe
///////////////////////////////////////////////////////////////////

void kfEnd()  {
  FFDControlPoints finalControlPoints;

  if (keyframeStart==false) return;         // no keyframe to end?
  if (absFrame) {                     // absolute x,y,z values given?
    finalControlPoints = offsetControlPoints;
  } else {                            // frame + relative offsets
    finalControlPoints = FFDadd(tempControlPoints, offsetControlPoints);
  }
  keyframeControlPoints.push_back(finalControlPoints);
  //  finalControlPoints.print("keyframe control points");
  keyframeStart = false;
}

///////////////////////////////////////////////////////////////////
// FUNC: loadScene()
// DOES: reads in scene as specifed by a text file
///////////////////////////////////////////////////////////////////

void loadScene()
{
  using namespace std;
  string line;
  int kfCount = 0;
 vector<string> words;

 absFrame = false;
 keyframeStart = false;
  ifstream myfile(sceneFile.c_str());          // open scene file
  if (!myfile.is_open()) {
    cerr << "Unable open scene: " << sceneFile << endl;
    return;
  }
  keyframeControlPoints.clear();            // reset all keyframes
  while(getParsedLine(myfile,words)) {      // for each line (non-blank, non-comment)
    string lineType = words[0];               // look at first word to determine line type
    if (lineType == "mesh") {                 // mesh to load
      mesh.readFromOBJ(words[1].c_str());
      if (words.size()==3) {             // optional size paramater for mesh
	float newsize = atof(words[2].c_str());
	mesh.normalize(newsize);
      }
      cout << "Read in mesh: " << words[1] << endl;
    }
    if (lineType == "bindframe") {    // bindframe
      double val[12];
      for (int n=0; n<12; n++) 
	val[n] = atof(words[n+1].c_str());   // read in 12 numbers
      bindFrame = Affine3(val);     // remember the bindframe
    }
    if (lineType == "keyframe")    kfStart(words);         // start of a keyframe
    if (lineType[0] == '(')        kfControlPoint(words);  // optional keyframe individual control points
    if (lineType == "endkeyframe") kfEnd();                // end of a keyframe
    if (lineType == "dtscale") dtScale = atof(words[1].c_str());                // end of a keyframe
  }
  kfEnd();         // end any ongoing keyframe specifications
  cout << "Read " << keyframeControlPoints.size() << " keyframes" << endl;
}

///////////////////////////////////////////////////////////////////
// FUNC: physicsInit()
// DOES: initialization for physics-based simulation of bottom layer of control points
///////////////////////////////////////////////////////////////////

void physicsInit() 
{
  switch(mode) {
  case 3:
    // Initialize positions
    deformedControlPoints = FFDControlPoints(Vector3(12,0,0), Vector3(0,11,0), Vector3(0,0,10), Vector3(-6,10,-5));
    //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
    //@               TO IMPLEMENT               @//
    //       Initialize velocities to zero

    //@                                          @//
    //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
    break;
  case 4:
    //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
    //@               TO IMPLEMENT               @//
    //     Initialize positions and velocities

    //@                                          @//
    //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
    break;
  }
}

///////////////////////////////////////////////////////////////////
// FUNC: physicsUpdate()
// DOES: time step update for physics-based simulation of bottom layer of control points
///////////////////////////////////////////////////////////////////

void physicsUpdate() 
{
  //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
  //@               TO IMPLEMENT               @//
  //         Use physics to animate the
  //       variable 'deformedControlPoints'

  //@                                          @//
  //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@//
}


